%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(distribution_SUITE).

%% Tests distribution and the tcp driver.

-include("test_server.hrl").

-export([all/1,
	 ping/1, bulk_send/1, bulk_send_small/1,
	 bulk_send_big/1,
	 local_send/1, local_send_small/1, local_send_big/1,
	 local_send_legal/1, link_to_busy/1, exit_to_busy/1,
	 lost_exit/1, link_to_dead/1, link_to_dead_new_node/1,
	 applied_monitor_node/1, ref_port_roundtrip/1,
	 trap_bif/1, trap_bif_1/1, trap_bif_2/1, trap_bif_3/1,
	 stop_dist/1, dist_auto_connect/1, 
	 dist_auto_connect_never/1, dist_auto_connect_once/1]).

%% Internal exports.
-export([sender/3, receiver2/2, dummy_waiter/0, dead_process/0,
	 roundtrip/1, do_dist_auto_connect/1]).

all(suite) -> [ping, bulk_send, local_send, link_to_busy, exit_to_busy,
	       lost_exit, link_to_dead, link_to_dead_new_node,
	       applied_monitor_node, ref_port_roundtrip, stop_dist, 
	       trap_bif, dist_auto_connect].

%%% Don't be too hard on vxworks, the cross server gets nodedown
%%% cause the card is too busy if we don't sleep a little between pings.
sleep() ->
    case os:type() of
	vxworks ->
	    receive
	    after 10 ->
		    ok
	    end;
	_ ->
	    ok
    end.

ping(doc) ->
    ["Tests pinging a node in different ways."];
ping(suite) -> [];
ping(Config) when list(Config) ->
    Times = 1024,

    %% Ping a non-existing node many times.  This used to crash the emulator
    %% on Windows.

    ?line Host = hostname(),
    ?line BadName = list_to_atom("__pucko__@" ++ Host),
    ?line io:format("Pinging ~s (assumed to not exist)", [BadName]),
    ?line test_server:do_times(Times, 
			       fun() -> pang = net_adm:ping(BadName),
					sleep() 
			       end),

    %% Pings another node.

    ?line {ok, OtherNode} = start_node(distribution_SUITE_other),
    ?line io:format("Pinging ~s (assumed to exist)", [OtherNode]),
    ?line test_server:do_times(Times, fun() -> pong = net_adm:ping(OtherNode),sleep() end),
    ?line stop_node(OtherNode),

    %% Pings our own node many times.

    ?line Node = node(),
    ?line io:format("Pinging ~s (the same node)", [Node]),
    ?line test_server:do_times(Times, fun() -> pong = net_adm:ping(Node),sleep() end),

    ok.

bulk_send(doc) ->
 ["Tests sending large amount of data to another node and measure",
  "the time. This tests that a process that is suspended on a ",
  "busy port will eventually be resumed."];
bulk_send(suite) ->
    [bulk_send_small, bulk_send_big].

bulk_send_small(suite) -> [];
bulk_send_small(Config) when list(Config) ->
    ?line bulk_send(64, 32, Config).

bulk_send_big(suite) -> [];
bulk_send_big(Config) when list(Config) ->
    ?line bulk_send(32, 64, Config).

bulk_send(Terms, BinSize, Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(30)),

    ?line io:format("Sending ~w binaries, each of size ~w K",
		    [Terms, BinSize]),
    ?line {ok, Node} = start_node(bulk_receiver),
    ?line Recv = spawn(Node, erlang, apply, [fun receiver/2, [0, 0]]),
    ?line Bin = list_to_binary(lists:duplicate(BinSize*1024, 253)),
    ?line Size = Terms*size(Bin),
    ?line {Elapsed, {Terms, Size}} = test_server:timecall(?MODULE, sender,
							  [Recv, Bin, Terms]),
    ?line stop_node(Node),

    ?line test_server:timetrap_cancel(Dog),
    {comment, integer_to_list(trunc(Size/1024/Elapsed+0.5)) ++ " K/s"}.

sender(To, Bin, 0) ->
    To ! {done, self()},
    receive
	Any ->
	    Any
    end;
sender(To, Bin, Left) ->
    To ! {term, Bin},
    sender(To, Bin, Left-1).

%% Receiver process to be run on a slave node.

receiver(Terms, Size) ->
    receive
	{term, Bin} ->
	    receiver(Terms+1, Size+size(Bin));
	{done, ReplyTo} ->
	    ReplyTo ! {Terms, Size}
    end.


local_send(suite) ->
    [local_send_small, local_send_big, local_send_legal];
local_send(doc) ->
    ["Tests sending small and big messages to a non-existing ",
     "local registered process."].

local_send_big(suite) -> [];
local_send_big(doc) ->
    ["Sends several big message to an non-registered process on ",
     "the local node."];
local_send_big(Config) when list(Config) ->
    Data0=local_send_big(doc)++local_send(doc),
    Data1=[Data0,[Data0, Data0, [Data0], Data0],Data0],
    Data2=Data0++lists:flatten(Data1)++
	list_to_binary(lists:flatten(Data1)),
    Func=fun() -> Data2= {arbitrary_name, node()} ! Data2 end,
    ?line test_server:do_times(4096, Func),
    ok.

local_send_small(suite) -> [];
local_send_small(doc) ->
    ["Sends a small message to an non-registered process on the ",
     "local node."];
local_send_small(Config) when list(Config) ->
    Data={some_stupid, "arbitrary", 'Data'},
    Func=fun() -> Data= {unregistered_name, node()} ! Data end,
    ?line test_server:do_times(4096, Func),
    ok.

local_send_legal(suite) -> [];
local_send_legal(doc) ->
    ["Sends data to a registered process on the local node, ",
     "as if it was on another node."];
local_send_legal(Config) when list(Config) ->
    Times=16384,
    Data={local_send_legal(doc), local_send_legal(doc)},
    Pid=spawn(?MODULE,receiver2, [0, 0]) ,
    ?line true=register(registered_process, Pid),

    Func=fun() -> Data={registered_process, node()} ! Data end,
    TotalSize=size(Data)*Times,
    ?line test_server:do_times(Times, Func),

    % Check that all msgs really came through.
    Me=self(),
    ?line {done, Me}=
	{registered_process, node()} ! {done, Me},
    receive
	{Times, TotalSize} ->
	    ok;
	_ ->
	    test_server:fail("Wrong number of msgs received.")
    end,
    ok.

receiver2(Num, TotSize) ->
    receive
	{done, ReplyTo} ->
	    ReplyTo ! {Num, TotSize};
	Stuff ->
	    receiver2(Num+1, TotSize+size(Stuff))
    end.

link_to_busy(doc) -> "Test that link/1 to a busy distribution port works.";
link_to_busy(suite) -> [];
link_to_busy(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(30)),
    ?line {ok, Node} = start_node(link_to_busy),
    ?line Recv = spawn(Node, erlang, apply, [fun sink/1, [link_to_busy_sink]]),
    %% Distribution delivers nodedown if all 64*1024 is sent to a VxWorks
    %% card during cross platform testing, we need to lower it a bit.
    %% Besides, we need to be careful with our limited memory...
    ?line Bin = case os:type() of
	vxworks ->
		?line list_to_binary(lists:duplicate(16*1024, 253));
	_ ->
		?line list_to_binary(lists:duplicate(64*1024, 253))
    end,

    %% We will spawn off a process which will try to link to the other
    %% node.  The linker process will not actually run until this
    %% process is suspended due to the busy distribution port (because
    %% of the big binary send).  When the link/1 is run, the linker
    %% process will block, too, because of the because busy port,
    %% and will later be restarted.

    ?line receive after 1 -> ok end,		% Clear reduction count.
    ?line fun_spawn(fun linker/1, [Recv]),
    ?line {link_to_busy_sink, Node} ! Bin,
    ?line {link_to_busy_sink, Node} ! Bin,

    %% Same thing, but we apply link/1 instead of calling it directly.

    ?line receive after 1 -> ok end,		% Clear reduction count.
    ?line fun_spawn(fun applied_linker/1, [Recv]),
    ?line {link_to_busy_sink, Node} ! Bin,
    ?line {link_to_busy_sink, Node} ! Bin,

    %% Same thing again, but we apply link/1 in the tail of a function.

    ?line receive after 1 -> ok end,		% Clear reduction count.
    ?line fun_spawn(fun tail_applied_linker/1, [Recv]),
    ?line {link_to_busy_sink, Node} ! Bin,
    ?line {link_to_busy_sink, Node} ! Bin,

    %% Done.
    ?line stop_node(Node),
    ?line test_server:timetrap_cancel(Dog),
    ok.

linker(Pid) ->
    true = link(Pid),
    {links, Links} = process_info(self(), links),
    true = lists:member(Pid, Links).

applied_linker(Pid) ->
    true = apply(erlang, link, [Pid]),
    {links, Links} = process_info(self(), links),
    true = lists:member(Pid, Links).

tail_applied_linker(Pid) ->
    apply(erlang, link, [Pid]).
    
exit_to_busy(doc) -> "Test that exit/2 to a busy distribution port works.";
exit_to_busy(suite) -> [];
exit_to_busy(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(30)),
    ?line {ok, Node} = start_node(exit_to_busy),
    ?line Bin = list_to_binary(lists:duplicate(64*1024, 253)),

    %% We will spawn off a process which will try to exit a process on
    %% the other node.  That process will not actually run until this
    %% process is suspended due to the busy distribution port (because
    %% of the big binary send).  The process executing exit/2 will block,
    %% too, because of the busy distribution port, and will be allowed
    %% to continue when the port becomes non-busy.

    ?line Recv1 = spawn(Node, erlang, apply, [fun sink/1, [exit_to_busy_sink]]),
    ?line receive after 1 -> ok end,		% Clear reduction count.
    ?line fun_spawn(fun joey_killer/1, [Recv1]),
    ?line {exit_to_busy_sink, Node} ! Bin,
    ?line {exit_to_busy_sink, Node} ! Bin,

    %% Same thing, but tail call to exit/2.
    ?line Recv2 = spawn(Node, erlang, apply, [fun sink/1, [exit_to_busy_sink]]),
    ?line receive after 1 -> ok end,		% Clear reduction count.
    ?line fun_spawn(fun tail_joey_killer/1, [Recv2]),
    ?line {exit_to_busy_sink, Node} ! Bin,
    ?line {exit_to_busy_sink, Node} ! Bin,

    %% Same thing, but we apply exit/2 instead of calling it directly.
    ?line Recv3 = spawn(Node, erlang, apply, [fun sink/1, [exit_to_busy_sink]]),
    ?line receive after 1 -> ok end,		% Clear reduction count.
    ?line fun_spawn(fun applied_joey_killer/1, [Recv3]),
    ?line {exit_to_busy_sink, Node} ! Bin,
    ?line {exit_to_busy_sink, Node} ! Bin,

    %% Same thing again, but we apply exit/2 in the tail of a function.
    ?line Recv4 = spawn(Node, erlang, apply, [fun sink/1, [exit_to_busy_sink]]),
    ?line receive after 1 -> ok end,		% Clear reduction count.
    ?line fun_spawn(fun tail_applied_joey_killer/1, [Recv4]),
    ?line {exit_to_busy_sink, Node} ! Bin,
    ?line {exit_to_busy_sink, Node} ! Bin,

    %% Give exit messages time to catch up with us.
    receive after 1000 -> ok end,
    
    %% Done.
    ?line stop_node(Node),
    ?line test_server:timetrap_cancel(Dog),
    ok.

remote_is_process_alive(Pid) ->
    rpc:call(node(Pid), erlang, is_process_alive,
	     [Pid]).

joey_killer(Pid) ->
    exit(Pid, die),
    receive after 100 -> ok end,
    false = remote_is_process_alive(Pid).

tail_joey_killer(Pid) ->
    exit(Pid, die).

applied_joey_killer(Pid) ->
    apply(erlang, exit, [Pid, die]),
    receive after 100 -> ok end,
    false = remote_is_process_alive(Pid).

tail_applied_joey_killer(Pid) ->
    apply(erlang, exit, [Pid, die]).

sink(Name) ->
    register(Name, self()),
    sink1().

sink1() ->
    receive
	Any -> sink1()
    end.

lost_exit(doc) ->
    "Test that EXITs messages send to another node are not lost if "
	"if the distribution port is busy.";
lost_exit(suite) -> [];
lost_exit(Config) when list(Config) ->
    ?line {ok, Node} = start_node(lost_exit),
    ?line Bin = list_to_binary(lists:duplicate(64*1024, 253)),
    ?line process_flag(trap_exit, true),

    %% Start a sink receiver on the other node.
    ?line Recv = spawn(Node, erlang, apply, [fun sink/1, [lost_exit_sink]]),

    %% Start one local and one remote process, and link them to each other.
    ?line Waiter = spawn_link(Node, ?MODULE, dummy_waiter, []),
    Self = self(),
    LFun =
	fun() ->
		link(Waiter),
		Self ! linked,
		receive
		    die ->
			exit(died)
		end
	end,
    ?line Linker = fun_spawn(LFun),

    %% Wait for the processes to be linked to each other, then send a
    %% message to the local process to kill itself.
    %% The message will not be received and acted on until this process
    %% blocks on a busy distribution port below.
   
    ?line receive 
	      linked -> ok;
	      Unexpected -> ?line ?t:fail({unpexected_message,Unexpected})
	  end,
    
    Linker ! die,
    ?line {lost_exit_sink, Node} ! Bin,
    ?line {lost_exit_sink, Node} ! Bin,
    ?line {lost_exit_sink, Node} ! Bin,

    %% Get rid of the 'EXIT' message from the local process.
    ?line receive {'EXIT', Linker, died} -> ok end,
    
    %% Now our messages are sent.  Therefore, the distribution port is no
    %% longer busy, and we will shortly receive an EXIT message from the
    %% remote process.

    ?line receive 
	      {'EXIT', Waiter, died} ->
		  ok;
	      Other ->
		  ?line test_server:fail({unexpected_message, Other})
	  after 10000 ->
		  ?line test_server:fail(process_still_alive)
	  end,

    %% Done.
    ?line stop_node(Node),
    ok.

dummy_waiter() ->
    receive
    after infinity ->
	    ok
    end.

link_to_dead(doc) ->
    ["Test that linking to a dead remote process gives an EXIT message ",
     "AND that the link is teared down."];
link_to_dead(suite) -> [];
link_to_dead(Config) when list(Config) ->
    ?line process_flag(trap_exit, true),
    ?line {ok, Node} = start_node(link_to_dead),
%    ?line monitor_node(Node, true),
    ?line net_adm:ping(Node), %% Ts_cross_server workaround.
    ?line Pid = spawn(Node, ?MODULE, dead_process, []),
    receive
    after 5000 -> ok
    end,
    ?line link(Pid),
    ?line receive
	      {'EXIT', Pid, noproc} ->
		  ok;
	      Other ->
		  ?line test_server:fail({unexpected_message, Other})
	  after 5000 ->
		  ?line test_server:fail(nothing_received)
	  end,
    ?line {links, Links} = process_info(self(), links),
    ?line io:format("Pid=~p, links=~p", [Pid, Links]),
    ?line false = lists:member(Pid, Links),
    ?line stop_node(Node),
    ?line receive
	      Message ->
		  ?line test_server:fail({unexpected_message, Message})
	  after 3000 ->
		  ok
	  end,
    ok.
    
dead_process() ->
    a = b.

link_to_dead_new_node(doc) ->
    ["Test that linking to a pid on node that has gone and restarted gives ",
     "the correct EXIT message (OTP-2304)."];
link_to_dead_new_node(suite) -> [];
link_to_dead_new_node(Config) when list(Config) ->
    ?line process_flag(trap_exit, true),

    %% Start the node, get a Pid and stop the node again.
    ?line {ok, Node} = start_node(link_to_dead_new_node),
    ?line Pid = spawn(Node, ?MODULE, dead_process, []),
    ?line stop_node(Node),

    %% Start a new node with the same name.
    ?line {ok, Node} = start_node(link_to_dead_new_node),
    ?line link(Pid),
    ?line receive
	      {'EXIT', Pid, noproc} ->
		  ok;
	      Other ->
		  ?line test_server:fail({unexpected_message, Other})
	  after 5000 ->
		  ?line test_server:fail(nothing_received)
	  end,

    %% Make sure that the link wasn't created.
    ?line {links, Links} = process_info(self(), links),
    ?line io:format("Pid=~p, links=~p", [Pid, Links]),
    ?line false = lists:member(Pid, Links),
    ?line stop_node(Node),
    ?line receive
	      Message ->
		  ?line test_server:fail({unexpected_message, Message})
	  after 3000 ->
		  ok
	  end,
    ok.

applied_monitor_node(doc) ->
    "Test that monitor_node/2 works when applied.";
applied_monitor_node(suite) -> [];
applied_monitor_node(Config) when list(Config) ->
    ?line NonExisting = list_to_atom("__non_existing__@" ++ hostname()),

    %% Tail-recursive call to apply (since the node is non-existing,
    %% there will be a trap).

    ?line true = tail_apply(erlang, monitor_node, [NonExisting, true]),
    ?line [{nodedown, NonExisting}] = test_server:messages_get(),

    %% Ordinary call (with trap).

    ?line true = apply(erlang, monitor_node, [NonExisting, true]),
    ?line [{nodedown, NonExisting}] = test_server:messages_get(),
    
    ok.

tail_apply(M, F, A) ->
    apply(M, F, A).

ref_port_roundtrip(doc) ->
    "Test that sending a port or reference to another node and back again "
	"doesn't correct them in any way.";
ref_port_roundtrip(suite) -> [];
ref_port_roundtrip(Config) when list(Config) ->
    ?line process_flag(trap_exit, true),
    ?line Port = open_port({spawn, efile}, []),
    ?line Ref = make_ref(),
    ?line {ok, Node} = start_node(ref_port_roundtrip),
    ?line net_adm:ping(Node),
    ?line Term = {Port, Ref},
    ?line io:format("Term before: ~p", [show_term(Term)]),
    ?line Pid = spawn_link(Node, ?MODULE, roundtrip, [Term]),
    ?line receive after 5000 -> ok end,
    ?line stop_node(Node),
    ?line receive
	      {'EXIT', Pid, {Port, Ref}} ->
		  ?line io:format("Term after: ~p", [show_term(Term)]),
		  ok;
	      Other ->
		  ?line io:format("Term after: ~p", [show_term(Term)]),
		  ?line test_server:fail({unexpected, Other})
	  after 10000 ->
		  ?line test_server:fail(timeout)
	  end,
    ok.

roundtrip(Term) ->
    exit(Term).

show_term(Term) ->
    binary_to_list(term_to_binary(Term)).

stop_dist(doc) ->
    ["Tests behaviour after net_kernel:stop (OTP-2586)."];
stop_dist(suite) -> [];
stop_dist(Config) when list(Config) ->
    ?line Str = os:cmd(atom_to_list(lib:progname())
		       ++ " -noshell -pa "
		       ++ ?config(data_dir, Config)
		       ++ " -s run"),
    ?line "true\n" = Str,
    {comment, "may fail on FreeBSD due to differently configured name lookup - ask arndt"}.

trap_bif(doc) ->
    ["Verifies that BIFs which are traps to Erlang work (OTP-2680)."];
trap_bif(suite) -> [trap_bif_1, trap_bif_2, trap_bif_3].

trap_bif_1(doc) ->
    [""];
trap_bif_1(suite) -> [];
trap_bif_1(Config) when list(Config) ->
    ?line {true} = tr1(),
    ok.

trap_bif_2(doc) ->
    [""];
trap_bif_2(suite) -> [];
trap_bif_2(Config) when list(Config) ->
    ?line {true} = tr2(),
    ok.

trap_bif_3(doc) ->
    [""];
trap_bif_3(suite) -> [];
trap_bif_3(Config) when list(Config) ->
    ?line {hoo} = tr3(),
    ok.

tr1() ->
    ?line NonExisting = 'abc@boromir',
    ?line X = erlang:monitor_node(NonExisting, true),
    {X}.

tr2() ->
    ?line NonExisting = 'abc@boromir',
    ?line X = apply(erlang, monitor_node, [NonExisting, true]),
    {X}.

tr3() ->
    ?line NonExisting = 'abc@boromir',
    ?line X = {NonExisting, glirp} ! hoo,
    {X}.



dist_auto_connect(doc) ->
    ["Tests the kernel parameter 'dist_auto_connect'."];
dist_auto_connect(suite) ->
    [dist_auto_connect_never, dist_auto_connect_once].


%% Start a relay node and a lonely (dist_auto_connect never) node.
%% Lonely node pings relay node. That should fail. 
%% Lonely node connects to relay node with net_kernel:connect_node/1.
%% Result is sent here through relay node.
dist_auto_connect_never(suite) -> [];
dist_auto_connect_never(Config) when list(Config) ->
    Self = self(),
    ?line {ok, RelayNode} = 
	start_node(dist_auto_connect_relay),
    ?line spawn(RelayNode, 
	      fun() ->
		      register(dist_auto_connect_relay, self()),
		      dist_auto_connect_relay(Self) 
	      end),
    ?line {ok, Handle} = dist_auto_connect_start(dist_auto_connect, never),
    ?line Result = 
	receive
	    {do_dist_auto_connect, ok} ->
		ok;
	    {do_dist_auto_connect, Error} ->
		{error, Error};
	    Other ->
		{error, Other}
	after 32000 ->
		timeout
	end,
    ?line stop_node(RelayNode),
    ?line Stopped =  dist_auto_connect_stop(Handle),
    ?line Junk = 
	receive 
	    {do_dist_auto_connect, _} = J -> 
		J 
	after 0 -> 
		ok 
	end,
    ?line {ok, ok, ok} = {Result, Stopped, Junk},
    ok.

%% Start a relay node and a lonely (dist_auto_connect once) node.
%% Lonely node sends a pid here through relay node.
%% We stop and start relay node, now without relay process.
%% We send a message (by pid) with our pid to lonely node 
%% that it can proceed.
%% Lonely node pings relay node. That should fail. 
%% Lonely node connects to relay node with net_kernel:connect_node/1.
%% Lonely node pings relay node. That should now succeed.
%% Result is sent here directly (by pid).
dist_auto_connect_once(suite) -> [];
dist_auto_connect_once(Config) when list(Config) ->
    Self = self(),
    ?line {ok, RelayNode} = 
	start_node(dist_auto_connect_relay),
    ?line spawn(RelayNode, 
	      fun() ->
		      register(dist_auto_connect_relay, self()),
		      dist_auto_connect_relay(Self) 
	      end),
    ?line {ok, Handle} = dist_auto_connect_start(dist_auto_connect, once),
    ?line Result =
	receive
	    {do_dist_auto_connect, Pid} when pid(Pid) ->
		?line stop_node(RelayNode),
		?line {ok, RelayNode} = 
		    start_node(dist_auto_connect_relay),
		?line Pid ! {dist_auto_connect_once, self()},
		?line receive
			  {do_dist_auto_connect, ok} ->
			      ok;
			  {do_dist_auto_connect, Error} ->
			      Error;
			  O ->
			      {error, O}
		      after 10000 ->
			      timeout
		      end;
	    Other ->
		{error, Other}
	after 32000 ->
		timeout
	end,
    ?line stop_node(RelayNode),
    ?line Stopped =  dist_auto_connect_stop(Handle),
    ?line Junk = 
	receive 
	    {do_dist_auto_connect, _} = J -> 
		J 
	after 0 -> 
		ok 
	end,
    ?line {ok, ok, ok} = {Result, Stopped, Junk},

    ok.


do_dist_auto_connect([never]) ->
    Node = list_to_atom("dist_auto_connect_relay@" ++ hostname()),
    io:format("~p:do_dist_auto_connect([false]) Node=~p~n", 
	      [?MODULE, Node]),
    Ping = net_adm:ping(Node),
    io:format("~p:do_dist_auto_connect([false]) Ping=~p~n", 
	      [?MODULE, Ping]),
    Result = case Ping of
		 pang -> ok;
		 _ -> {error, Ping}
	     end,
    io:format("~p:do_dist_auto_connect([false]) Result=~p~n", 
	      [?MODULE, Result]),
    net_kernel:connect_node(Node),
    catch {dist_auto_connect_relay, Node} ! {do_dist_auto_connect, Result};
%    receive after 1000 -> ok end,
%    halt();
do_dist_auto_connect([once]) ->
    Node = list_to_atom("dist_auto_connect_relay@" ++ hostname()),
    io:format("~p:do_dist_auto_connect([false]) Node=~p~n", 
	      [?MODULE, Node]),
    catch {dist_auto_connect_relay, Node} ! {do_dist_auto_connect, self()},
    Parent = receive {dist_auto_connect_once, P} when pid(P) -> P end,
    Ping1 = net_adm:ping(Node),
    io:format("~p:do_dist_auto_connect([false]) Ping1=~p~n", 
	      [?MODULE, Ping1]),
    net_kernel:connect_node(Node),
    Ping2 = net_adm:ping(Node),
    io:format("~p:do_dist_auto_connect([false]) Ping2=~p~n", 
	      [?MODULE, Ping2]),
    Result = case Ping = {Ping1, Ping2} of
		 {pang, pong} -> ok;
		 _ -> {error, Ping}
	     end,
    io:format("~p:do_dist_auto_connect([false]) Result=~p~n", 
	      [?MODULE, Result]),
    catch Parent ! {do_dist_auto_connect, Result};
%    receive after 1000 -> ok end,
%    halt();
do_dist_auto_connect(Arg) ->
    io:format("~p:do_dist_auto_connect(~p)~n", 
	      [?MODULE, Arg]),
    receive after 10000 -> ok end,
    halt().
	    

dist_auto_connect_start(Name, Value) when atom(Name) ->
    dist_auto_connect_start(atom_to_list(Name), Value);
dist_auto_connect_start(Name, Value) when list(Name), atom(Value) ->
    Node = list_to_atom(lists:append([Name, "@", hostname()])),
    ModuleDir = filename:dirname(code:which(?MODULE)),
    ValueStr = atom_to_list(Value),
    Cookie = atom_to_list(erlang:get_cookie()),
    Cmd = lists:concat(
	    [%"xterm -e ",
	     atom_to_list(lib:progname()),
%	     " -noinput ",
	     " -detached ", 
	     long_or_short(), " ", Name,
	     " -setcookie ", Cookie,
	     " -pa ", ModuleDir,
	     " -s ", atom_to_list(?MODULE), 
	     " do_dist_auto_connect ", ValueStr,
	     " -kernel dist_auto_connect ", ValueStr]),
    io:format("~p:dist_auto_connect_start() cmd: ~p~n", [?MODULE, Cmd]),
    Port = open_port({spawn, Cmd}, [stream]),
    {ok, {Port, Node}}.


dist_auto_connect_stop({Port, Node}) ->
    Pid = spawn_link(fun() -> rpc:call(Node, erlang, halt, []) end),
    dist_auto_connect_stop(Port, Node, Pid, 5000).

dist_auto_connect_stop(Port, _Node, Pid, N) when integer(N), N =< 0 ->
    exit(Pid, normal),
    catch erlang:port_close(Port),
    Result = {error, node_not_down},
    io:format("~p:dist_auto_connect_stop() ~p~n", [?MODULE, Result]),
    Result;
dist_auto_connect_stop(Port, Node, Pid, N) when integer(N) ->
    case net_adm:ping(Node) of
	pong ->
	    receive after 100 -> ok end,
	    dist_auto_connect_stop(Port, Node, Pid, N-100);
	pang ->
	    exit(Pid, normal),
	    catch erlang:port_close(Port),
	    io:format("~p:dist_auto_connect_stop() ok~n", [?MODULE]),
	    ok
    end.


dist_auto_connect_relay(Parent) ->	    
    receive X ->
	    catch Parent ! X
    end,
    dist_auto_connect_relay(Parent).

%%% Utilities

start_node(Name) ->
    start_node(Name, []).

start_node(Name, Args) ->
    Pa = filename:dirname(code:which(?MODULE)),
    Cookie = atom_to_list(erlang:get_cookie()),
    test_server:start_node(Name, slave, 
			   [{args,
			     Args++" -setcookie "++Cookie++" -pa "++Pa}]).

stop_node(Node) ->
    test_server:stop_node(Node).

hostname() ->
    from($@, atom_to_list(node())).

from(H, [H | T]) -> T;
from(H, [_ | T]) -> from(H, T);
from(H, []) -> [].

fun_spawn(Fun) ->
    fun_spawn(Fun, []).

fun_spawn(Fun, Args) ->
    spawn_link(erlang, apply, [Fun, Args]).


long_or_short() -> 
    case net_kernel:longnames() of
	true -> " -name ";
	false -> " -sname "
    end.

