%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(op_SUITE).

-include("test_server.hrl").

-export([all/1, logical/1, t_not/1, relop/1, complex_relop/1]).

-export([init_per_testcase/2, fin_per_testcase/2]).
-import(lists, [foldl/3,flatmap/2]).

all(suite) -> [logical, t_not, relop, complex_relop].

init_per_testcase(Func, Config) when atom(Func), list(Config) ->
    Dog=?t:timetrap(?t:minutes(3)),
    [{watchdog, Dog}|Config].

fin_per_testcase(Func, Config) ->
    Dog=?config(watchdog, Config),
    ?t:timetrap_cancel(Dog).


logical(doc) -> "Test the logical operators and internal BIFs.";
logical(suite) -> [];
logical(Config) ->
    Vs0 = [true,false,bad],
    Vs = [unvalue(V) || V <- Vs0],
    Cases = [{Op,X,Y} || Op <- ['and','or','xor'], X <- Vs, Y <- Vs],
    ?line run_test_module(Cases, false),
    {comment,integer_to_list(length(Cases)) ++ " cases"}.

t_not(doc) -> "Test the not operator and internal BIFs.";
t_not(suite) -> [];
t_not(Config) when list(Config) ->
    ?line Cases = [{'not',unvalue(V)} || V <- [true,false,42,bad]],
    ?line run_test_module(Cases, false),
    {comment,integer_to_list(length(Cases)) ++ " cases"}.

relop(doc) -> "Test the relational operators and internal BIFs on literals.";
relop(suite) -> [];
relop(Config) when list(Config) ->
    Big1 = -38374938373887374983978484,
    Big2 = 19738924729729787487784874,
    F1 = float(Big1),
    F2 = float(Big2),
    Vs0 = [a,b,-33,-33.0,0,0.0,42,42.0,Big1,Big2,F1,F2],
    ?line Vs = [unvalue(V) || V <- Vs0],
    Ops = ['==', '/=', '=:=', '=/=', '<', '=<', '>', '>='],
    ?line binop(Ops, Vs).

complex_relop(doc) ->
    "Test the relational operators and internal BIFs on lists and tuples.";
complex_relop(suite) -> [];
complex_relop(Config) when list(Config) ->
    Big = 99678557475484872464269855544643333,
    Float = float(Big),
    Vs0 = [an_atom,42.0,42,Big,Float],
    Vs = flatmap(fun(X) -> [unvalue({X}),unvalue([X])] end, Vs0),
    Ops = ['==', '/=', '=:=', '=/=', '<', '=<', '>', '>='],
    ?line binop(Ops, Vs).

binop(Ops, Vs) ->
    Run = fun(Op, N) -> ?line Cases = [{Op,V1,V2} || V1 <- Vs, V2 <- Vs],
			?line run_test_module(Cases, true),
			N + length(Cases) end,
    ?line NumCases = foldl(Run, 0, Ops),
    {comment,integer_to_list(NumCases) ++ " cases"}.
    
run_test_module(Cases, GuardsOk) ->
    ?line Es = [expr(C) || C <- Cases],
    ?line Ok = unvalue(ok),
    ?line Gts = case GuardsOk of
		    true ->
			?line lists:foldr(fun guard_test/2, [Ok], Es);
		    false ->
			[unvalue(ok)]
		end,
    ?line Fun1 = make_function(guard_tests, Gts),
    ?line Bts = lists:foldr(fun body_test/2, [Ok], Es),
    ?line Fun2 = make_function(body_tests, Bts),
    ?line Bbts = lists:foldr(fun internal_bif/2, [Ok], Es),
    ?line Fun3 = make_function(bif_tests, Bbts),
    ?line Module = make_module(op_tests, [Fun1,Fun2,Fun3]),
    ?line lists:foreach(fun(F) -> io:put_chars([erl_pp:form(F),"\n"]) end, Module),

    %% Compile, load, and run the generated module.

    ?line {ok,Mod,Code1} = compile:forms(Module, [time]),
    ?line code:delete(Mod),
    ?line code:purge(Mod),
    ?line {module,Mod} = code:load_binary(Mod, Mod, Code1),
    ?line run_function(Mod, guard_tests),
    ?line run_function(Mod, body_tests),
    ?line run_function(Mod, bif_tests),
    ?line true = code:delete(Mod),
    ?line code:purge(Mod),

    ok.

expr({Op,X}) ->
    E = {op,1,Op,X},
    Res = eval([E]),
    {E,{Op,X},Res};
expr({Op,X,Y}) ->
    E = {op,1,Op,X,Y},
    Res = eval([E]),
    {E,{Op,value(X),value(Y)},Res}.

run_function(Mod, Name) ->
    case catch Mod:Name() of
	{'EXIT',Reason} ->
	    io:format("~p", [get(last)]),
	    ?t:fail({'EXIT',Reason});
	Other ->
	    ok
    end.
    
guard_test({E,Expr,Res}, Tail) ->
    True = unvalue(true),
    [save_term(Expr),
     {match,1,unvalue(Res),
      {'if',1,[{clause,1,[],[[E]],[True]},
	       {clause,1,[],[[True]],[unvalue(false)]}]}}|Tail].

body_test({E,Expr,{'EXIT',_}}, Tail) ->
    [save_term(Expr),
     {match,1,{tuple,1,[unvalue('EXIT'), {var,1,'_'}]},
      {'catch',1,E}}|Tail];
body_test({E,Expr,Res}, Tail) ->
    [save_term(Expr),
     {match,1,unvalue(Res),E}|Tail].

internal_bif({{op,_,Op,X}=E,Expr,Res}, Tail) ->
    internal_bif(Op, [X], Expr, Res, Tail);
internal_bif({{op,_,Op,X,Y}=E,Expr,Res}, Tail) ->
    internal_bif(Op, [X,Y], Expr, Res, Tail).

internal_bif(Op, Args, Expr, {'EXIT',_}, Tail) ->
    [save_term(Expr),
     {match,1,{tuple,1,[unvalue('EXIT'), {var,1,'_'}]},
      {'catch',1,{call,1,{remote,1,{atom,1,erlang},unvalue(Op)},Args}}}|Tail];
internal_bif(Op, Args, Expr, Res, Tail) ->
    [save_term(Expr),
     {match,1,unvalue(Res),
      {call,1,{remote,1,{atom,1,erlang},unvalue(Op)},Args}}|Tail].

save_term(Term) ->
    {call,1,
       {atom,1,put},
       [{atom,1,last},unvalue(Term)]}.

make_module(Name, Funcs) ->
    [{attribute,1,module,Name},
     {attribute,0,compile,export_all}|Funcs ++ [{eof,0}]].

make_function(Name, Body) ->
    {function,1,Name,0,[{clause,1,[],[],Body}]}.
       
eval(E) ->
    ?line case catch erl_eval:exprs(E, []) of
	      {'EXIT',Reason} -> {'EXIT',Reason};
	      {value,Val,Bs} -> Val
	  end.

unvalue(V) -> erl_parse:abstract(V).
    
value({nil,_}) -> [];
value({integer,_,X}) -> X;
value({string,_,X}) -> X;
value({float,_,X})   -> X;
value({atom,_,X})    -> X;
value({tuple,_,Es}) ->
    list_to_tuple(lists:map(fun(X) -> value(X) end, Es));
value({cons,_,H,T}) ->
    [value(H) | value(T)].
