/* Sl alloc implementation */

#define HARD_DEBUG 1

#if defined(INCLUDE_NEW_SL_ALLOC)
#  include "erl_sl_alloc.c"
#elif defined(INCLUDE_OLD_SL_ALLOC)
#  include "erl_old_sl_alloc.c"
#else
#  error "Include which sl_alloc?"
#endif

#define SUCCESS_EXIT_CODE 0
#define FAIL_EXIT_CODE    1
#define SKIP_EXIT_CODE    17

#define PRINT_ALLOC_OPS

/* Some functions which makes it possible to build without bringing in
   the whole emulator */

#ifdef INSTRUMENT

void *sys_alloc2(Uint s)		{ return malloc(s);                  }
void *sys_realloc2(void *p, Uint s)	{ return realloc(p, s);              }
void  sys_free2(void *p)		{        free(p);                    }


void *sys_alloc(Uint s)			{ return sys_alloc2(s);              }
void *sys_realloc(void *p, Uint s)	{ return sys_realloc2(p, s);         }
void  sys_free(void *p)			{        sys_free2(p);               }


void *erts_sl_alloc(Uint s)		{ return erts_sl_alloc2(s);          }
void *erts_sl_realloc(void *p,
		      Uint s1,
		      Uint s2)		{ return erts_sl_realloc2(p, s1, s2);}
void  erts_sl_free(void *p)		{        erts_sl_free2(p);           }

void *instr_alloc(int from,
		  void *(*alloc_func)(Uint),
		  Uint size)		{ return (*alloc_func)(size);        }

#else

void *sys_alloc(Uint s)			{ return malloc(s);                  }
void *sys_realloc(void *p,
		  Uint s)		{ return realloc(p, s);              }
void  sys_free(void *p)			{        free(p);                    }

#endif


#ifdef DEBUG
void erl_assert_error(char* expr, char* file, int line)
{   
    fflush(stdout);
    printf("Assertion failed: %s in %s, line %d\n",
            expr, file, line);
    fflush(stdout);
    abort();
}
#endif

void erl_error(char *fmt, va_list args)
{
    vfprintf(stderr, fmt, args);
}

void erl_exit(int n, char *fmt,...)
{
    va_list args;
    va_start(args, fmt);
    erl_error(fmt, args);
    va_end(args);
    exit(1);
}

void sys_printf(CIO where, char* format, ...)
{
    va_list va;
    va_start(va,format);

    if(where != CERR && where != COUT)
      where = CERR;

    if (where == CERR)
	erl_error(format, va);
    else
        vfprintf(stdout, format, va);
      
    va_end(va);
}

static int  test_new_sl_alloc;
static char empty_string[] = "";
static char *progname;
static char *argument;

/* Some convenience makros and functions */

#define START_TEST							\
int main(int argc, char *argv[])					\
{									\
    progname = empty_string;						\
    argument = empty_string;						\
    if (argc > 0)							\
	progname = argv[0];						\
    if (argc > 1) {							\
	if (strcmp("new_sl_alloc", argv[1]) == 0)			\
	    test_new_sl_alloc = 1;					\
	else if (strcmp("old_sl_alloc", argv[1]) == 0)			\
	    test_new_sl_alloc = 0;					\
	else {								\
	    printf("%s: Don't know which sl_alloc to test\n",		\
		   __FILE__);						\
	    fflush(stdout);						\
	    exit(FAIL_EXIT_CODE);					\
	}								\
    }									\
    if(argc > 2)							\
	argument = argv[2];						\
    if (argc > 3) {							\
	printf("%s: Test case called with too many arguments\n",	\
	       __FILE__);						\
	fflush(stdout);							\
	exit(FAIL_EXIT_CODE);						\
    }

#define END_TEST							\
    TEST_SUCCEEDED;							\
}

#define FAIL_TEST                                                           \
do {                                                                        \
  printf("%s:%d: Test case %s failed!\n", __FILE__, __LINE__, progname);    \
  fflush(stdout); abort();                                                  \
  exit(FAIL_EXIT_CODE);                                                     \
} while (0)

#define SKIP_TEST(S)                                                        \
do {                                                                        \
  printf("%s:%d: Test case %s skipped: %s\n",                               \
	 __FILE__,                                                          \
	 __LINE__,                                                          \
	 progname,                                                          \
	 (S));                                                              \
  fflush(stdout);                                                           \
  exit(SKIP_EXIT_CODE);                                                     \
} while (0)

#define TEST_SUCCEEDED                                                      \
do {                                                                        \
  printf("%s:%d: Test case %s succeeded!\n", __FILE__, __LINE__, progname); \
  fflush(stdout);                                                           \
  exit(SUCCESS_EXIT_CODE);                                                  \
} while (0)


typedef struct {
  byte *p;
  Uint  s;
  int   i;
  Uint *as;
} block;


#define CHECK_BLOCK_DATA(P, S, D)                                            \
do {                                                                         \
  int i__;                                                                   \
  for(i__ = 0; i__ < (S); i__++)                                             \
    if((P)[i__] != (byte) (D)) {                                             \
      printf("Data clobbered! found id=%d; expected id=%d\n",                \
	     (int) (P)[i__], (int) (D));                                     \
      FAIL_TEST;                                                             \
    }                                                                        \
} while (0)


#define SET_BLOCK_DATA(P, S, D)                                              \
do {                                                                         \
  int i__;                                                                   \
  for(i__ = 0; i__ < (S); i__++)                                             \
    ((byte *) (P))[i__] = (byte)(D);                                         \
} while (0)

void
alloc_op(block *bp, int id, int clean_up)
{
  if(bp->p)
    CHECK_BLOCK_DATA(bp->p, bp->s, id);

  if(bp->as[bp->i] == 0 || clean_up) {
    erts_sl_free(bp->p);
#ifdef PRINT_ALLOC_OPS
    printf("erts_sl_free(%p) [id=%d]\n", bp->p, id);
    fflush(stdout);
#endif
    bp->p = NULL;
    bp->s = 0;
    bp->i = 0; /* start from the beginning again */
    return;
  }

  if(!bp->p) {
    bp->s = bp->as[bp->i];
    bp->p = (byte *) erts_sl_alloc(bp->s);
#ifdef PRINT_ALLOC_OPS
    printf("%p = erts_sl_alloc(%u) [id=%d]\n",
	   bp->p, (unsigned int) bp->s, id);
    fflush(stdout);
#endif
    if(!bp->p) {
      printf("erts_sl_alloc(%u) failed [id=%d])\n",
	     (unsigned int) bp->s, id);
      FAIL_TEST;
    }
    SET_BLOCK_DATA(bp->p, bp->s, id);
  }
  else {
    byte *p = (byte *) erts_sl_realloc(bp->p, bp->s, bp->as[bp->i]);
#ifdef PRINT_ALLOC_OPS
    printf("%p = erts_sl_realloc(%p, %u, %u) [id=%d]\n",
	   p, bp->p, (unsigned int) bp->s, (unsigned int) bp->as[bp->i], id);
    fflush(stdout);
#endif
    if(!p) {
      printf("erts_sl_realloc(%p, %u, %u) failed [id=%d]\n",
	     bp->p,
	     (unsigned int) bp->s,
	     (unsigned int) bp->as[bp->i],
	     id);
      FAIL_TEST;
    }

    if(bp->s < bp->as[bp->i]) {
      CHECK_BLOCK_DATA(p, bp->s, id);
      SET_BLOCK_DATA(p, bp->as[bp->i], id);
    }
    else
      CHECK_BLOCK_DATA(p, bp->as[bp->i], id);

    bp->s = bp->as[bp->i];
    bp->p = p;
  }

  bp->i++;
}
